# Procs to support testing OpenACS with Tclwebtest.
#
# Basic utility procs.
#
# @author Peter Marklund

namespace eval ::twt {}

ad_proc ::twt::log_section { message } {
    set script_name [file tail [info script]]
    puts ""
    puts "##############################"
    puts "#"
    puts "# ${script_name}: $message"
    puts "#"
    puts "##############################"
    puts ""
}

ad_proc ::twt::log { message } {
    set script_name [file tail [info script]]
    puts "${script_name}: $message"
}

ad_proc ::twt::log_alert { message } {
    set script_name [file tail [info script]]
    puts ""
    puts "${script_name}: [::twt::config::alert_keyword] - $message"
    puts ""
}

ad_proc ::twt::assert { explanation expression } {
    if { !$expression } {
        ::twt::log_alert "Assertion \"$explanation\" failed"
    }
} 

ad_proc ::twt::assert_equals { explanation actual_value expected_value } {
    if { ![string equal $actual_value $expected_value] } {
        ::twt::log_alert "Assertion \"$explanation\" failed: actual_value=\"$actual_value\", expected_value=\"$expected_value\""
    }
} 

ad_proc ::twt::do_request { page_url } {
    Takes a a url and invokes tclwebtest::do_request. Will retry
    the request a number of times if it fails because of a socket
    connect problem.
} {
    # Qualify page_url if necessary
    if { [regexp {^/} $page_url] } {
        set page_url "[::twt::config::server_url]${page_url}"
    }

    set retry_count 0
    set retry_max 10
    set error_p 0
    while { [catch {::tclwebtest::do_request $page_url} errmsg] } {
        set error_p 1

        if { $retry_count < $retry_max } {
            switch -regexp -- $errmsg {
                {unreachable} - {refused} {
                    ::twt::log "Failed to connect to server with error \"$errmsg\" - retrying"
                    incr retry_count
                    exec "sleep" "5"
                    set error_p 0
                    continue
                }
                default {
                    ::twt::log "Failed to connect to server with error \"$errmsg\" - giving up"
                    break
                }
            }
        } else {
            break
        }
    }

    if { $error_p } {
        # Either some non-socket error, or a socket problem occuring with more than
        # $retry_max times. Propagate the error while retaining the stack trace
        global errorInfo
        error $errmsg $errorInfo
    }

    ::twt::acs_lang::check_no_keys
}

ad_proc ::twt::get_url_list { page_url link_url_pattern } {

    ::twt::do_request $page_url

    set urls_list [list]

    # Loop over and add all links
    set errno "0"
    while { $errno == "0" } {
	set errno [catch {
            array set link_array [link find -next ~u "$link_url_pattern"]} error]

         if { [string equal $errno "0"] } {
            set url $link_array(url)
     
            lappend urls_list $url
        }
    }
    

    return $urls_list
}

ad_proc ::twt::get_random_items_from_list { list number } {
    Given a list and the lenght of the list to return, 
    return a list with a random subset of items from the input list.
} {

    # Build a list of indices
    set index_list [list]
    for { set i 0 } { $i < [llength $list] } { incr i } {
        lappend index_list $i
    }

    # If the list was empty - return
    if { [llength $index_list] == 0 } {
        return {}
    }

    # Cannot return more items than are in the list
    if { $number > [llength $list] } {
        error "get_random_items_from_list: cannot return $number items from list $list"
    }

    # Pick number random indices from the list. Remove each index that we have
    # already picked.
    set random_indices [list]
    for { set index_count 0 } { $index_count < $number } { incr index_count } {
        set random_index [randomRange [llength $index_list]]

        lappend random_indices [lindex $index_list $random_index]

        # Remove the index that we picked
        set index_list [lreplace $index_list $random_index $random_index]
    }

    # Build and return the items at the random indices
    set return_list [list]
    foreach index $random_indices {
        lappend return_list [lindex $list $index]
    }
    if { [llength $return_list] == 1 } {
        return [lindex $return_list 0]
    } else {
        return $return_list
    }
}

ad_proc ::twt::randomRange {range} {
    Given an integer N, return an integer between 0 and N.
} {

     return [expr int([expr rand()] * $range)]
}

ad_proc ::twt::write_response_to_file { filename } {
    Write the HTML body of the last HTTP response to the
    file with given path. If the directory of the file doesn't
    exist then create it.
} {

    # Create the directory of the output file if it doesn't exist
    if { ![file isdirectory [file dirname $filename]] } { 
        exec mkdir -p [file dirname $filename] 
    }
    set file_id [open "$filename" w+]
    puts $file_id "[response body]"
}

ad_proc ::twt::crawl_links {} {
    TODO: This proc currently doesn't crawl nearly as many links as we would like
} {

    global __url_history

    set start_url [lindex $__url_history end]

    # Return if given start URL is external
    global __server_url
    set absolute_url [tclwebtest::absolute_link $start_url]
    if { [string first $__server_url $absolute_url] == -1 } {
        return
    }

    # Also return if this is the logout link
    if { [regexp {/register/logout} $start_url match] } {
        return
    }

    ::twt::do_request $start_url

    set errno 0
    while { $errno == "0" } {
	set errno [catch {
            array set link_array [link find -next]} error]

         if { [string equal $errno "0"] } {
            set url $link_array(url)

            # Don't revisit URL:s we have already tested
            # Don't follow relative anchors on pages - can't get them to work with TclWebtest
            if { [lsearch -exact $__url_history $url] == -1 && [string range $url 0 0] != "#" } {
                lappend __url_history $url

                crawl_links
            } else {
            }
         }
   }
}

ad_proc ::twt::multiple_select_value { name value } {
    Selects the option with the given value in the current
    form widget (workaround since I can only get tclwebtest
    to select based on label).
} {
    field find ~n $name

    array set current_field [field current]
    set field_choices $current_field(choices)
    set index 0
    foreach field_choice $field_choices {
        if { [string equal $value [lindex $field_choice 0]] } {
            break
        }
        incr index
    }

    ::tclwebtest::field_select -index $index
}

ad_proc ::twt::count_links { pattern } {

    set count 0
    foreach link_list [link all] {
        array set link $link_list

        if { [regexp $pattern $link(url)] } {
            incr count
        }
    }

    return $count
}

# No longer needed as the problem was due to malformed HTML and is
# better handled within tclwebtest
#  ad_proc ::twt::all_urls {} {
#      Returns all urls on the page last requested.

#      This proc is a workaround to the problem
#      with Tclwebtest not always finding all links with the [link all]
#      command.

#      @return A list of URLs of the last requested page

#      @author Peter Marklund
#  } {
#      set all_urls [list]
#      set remaining_body [response body]
#      while { [regexp -all {^(.*?)<a href="?([^">]+)(.*)$} $remaining_body match before_match link_url remaining_body] } {
#          lappend all_urls $link_url
#      }    
    
#      return $all_urls
#  }
